/*
Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA.
All rights reserved.

This Agreement gives you, the LICENSEE, certain rights and obligations.
By using the software, you indicate that you have read, understood, and
will comply with the terms.

Permission to use, copy and modify for internal, noncommercial purposes
is hereby granted.  Any distribution of this program or any part thereof
is strictly prohibited without prior written consent of M.I.T.

Title to copyright to this software and to any associated documentation
shall at all times remain with M.I.T. and LICENSEE agrees to preserve
same.  LICENSEE agrees not to make any copies except for LICENSEE'S
internal noncommercial use, or to use separately any portion of this
software without prior written consent of M.I.T.  LICENSEE agrees to
place the appropriate copyright notice on any such copies.

Nothing in this Agreement shall be construed as conferring rights to use
in advertising, publicity or otherwise any trademark or the name of
"Massachusetts Institute of Technology" or "M.I.T."

M.I.T. MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  By
way of example, but not limitation, M.I.T. MAKES NO REPRESENTATIONS OR
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR
THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS OR DOCUMENTATION WILL
NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
M.I.T. shall not be held liable for any liability nor for any direct,
indirect or consequential damages with respect to any claim by LICENSEE
or any third party on account of or arising from this Agreement or use
of this software.
*/
/* This program wil generate input files with constant-type 
 * elements (quads and triangles) for a simplke case.
 * At the moment most of the stuff is hardcoded. 
 * Eventually, most should be reached through command line arguments */

# include <stdlib.h>
# include <stdio.h>
# include <math.h>
# include <string.h>
# include <assert.h>

#define PI 3.141592653589793
#define ABS(A) ( ( (A) > 0 ) ? (A) : (-(A)) )

int main(int argc, char *argv[])
{
  int nDense; /* Higher value = denser grid */
  int i,j, iside, ic, iplot, swapNormal, gradeMesh[6];
  double sideLength, center[3];
  double dx,du, u[4],v[4], xout[4],yout[4],zout[4],
    xu,xv,x0, yu,yv,y0, zu,zv,z0, *uvals, *ugvals, *uuse;
  char inStr[]="inwards", outStr[]="outwards", *normalStr;
  FILE *outFile;

  outFile = fopen("cube.qui","w");

  /* Generate a cube with equal size mesh and outward pointing normal */
  sideLength=1;
  center[0]=0.5;
  center[1]=0.5;
  center[2]=0.5;
  /* nDense by nDense elements on each side. */
  nDense = 32;
  /* Possibly swap normal to point inwards: */
  swapNormal=0;
  /* Possibly grade the mesh using a cosine scaling towards the edges */
  gradeMesh[0] = 0; /* normal = ( 1, 0, 0);*/
  gradeMesh[1] = 0; /* normal = (-1, 0, 0);*/
  gradeMesh[2] = 0; /* normal = ( 0, 1, 0);*/
  gradeMesh[3] = 0; /* normal = ( 0,-1, 0);*/
  gradeMesh[4] = 0; /* normal = ( 0, 0, 1); "free surface"*/
  gradeMesh[5] = 0; /* normal = ( 0, 0,-1);*/

  /* Plot # elements to be generated: */
  fprintf(stdout,"Generating %d elements\n",6*nDense*nDense);

  /* Mesh size (if uniform) */
  dx = sideLength/(double) nDense;


  /* Setup number of different coordinate values to be used, 
   * and set those values */
  uvals = (double *) calloc(nDense+1,sizeof(double));
  ugvals = (double *) calloc(nDense+1,sizeof(double));
  du = 2.0/(double) nDense;
  /* Points located from [-1:1] */
  uvals[0] = -1.0;
  for (i=1; i<nDense; i++)uvals[i] = i*du - 1.0;
  uvals[nDense] = 1.0;

  /* Cosine scaling of the vector */
  for (i=0; i<nDense+1; i++)
    ugvals[i] = cos(0.5*PI*(uvals[i]-1.0));
  /* Scale mesh with sidelength */
  for (i=0; i<nDense+1; i++) {
    uvals[i] *= sideLength/2.0;
    ugvals[i] *= sideLength/2.0;
  }

  normalStr = outStr;
  if (swapNormal) normalStr = inStr;
  /* Write header: */
  fprintf(outFile,
	  "0 Cube w/ %s pointing normal. %d by %d elements on each side\n",
	  normalStr,nDense,nDense);
  fprintf(outFile,
	  "# This file was automatically generated by \"cubegen\"\n");

  /* if I can set up ONE side, then the rest can be found from 
   * rotations (and possible translations), but no mirroring(!) of 
   * the first side. */
  for (iside=0; iside<6; iside++){
    switch (iside){
      /* Note: (u,v,n) is a right-hand-system. (outward normal) */
    case 0: 
      yu=1.0; yv=0.0; y0=0.0;
      zu=0.0; zv=1.0; z0=0.0;
      xu=0.0; xv=0.0; x0=sideLength/2.0;
      break;
    case 1: 
      yu=1.0; yv=0.0; y0=0.0;
      zu=0.0; zv=-1.0; z0=0.0;
      xu=0.0; xv=0.0; x0=-sideLength/2.0;
      break;
    case 2: 
      xu=1.0; xv=0.0; x0=0.0;
      zu=0.0; zv=-1.0; z0=0.0;
      yu=0.0; yv=0.0; y0=sideLength/2.0;
      break;
    case 3: 
      xu=1.0; xv=0.0; x0=0.0;
      zu=0.0; zv=1.0; z0=0.0;
      yu=0.0; yv=0.0; y0=-sideLength/2.0;
      break;
    case 4: 
      xu=1.0; xv=0.0; x0=0.0;
      yu=0.0; yv=1.0; y0=0.0;
      zu=0.0; zv=0.0; z0=sideLength/2.0;
      break;
    case 5: 
      xu=1.0; xv=0.0; x0=0.0;
      yu=0.0; yv=-1.0; y0=0.0;
      zu=0.0; zv=0.0; z0=-sideLength/2.0;
      break;
    }
    /* Check if we should grade mesh on this side: */
    if (gradeMesh[iside])uuse = ugvals; /* Graded mesh      */
    else uuse = uvals;                  /* Equidistant mesh */
    for (i=0; i<nDense; i++){
      for (j=0; j<nDense; j++){
	/* Coordinate for this element in u,v;
	 * normal upwards, clockwise ordering  */
	u[0] = uuse[i]; 
	v[0] = uuse[j]; 
	u[1] = uuse[i]; 
	v[1] = uuse[j+1]; 
	u[2] = uuse[i+1]; 
	v[2] = uuse[j+1]; 
	u[3] = uuse[i+1]; 
	v[3] = uuse[j]; 

	/* For each corner:*/ 
	for (ic=0; ic<4; ic++){
	/* Correlate the (u,v) to (x,y,z) for the given side */
	  xout[ic] = xu*u[ic] +xv*v[ic] +x0;
	  yout[ic] = yu*u[ic] +yv*v[ic] +y0;
	  zout[ic] = zu*u[ic] +zv*v[ic] +z0;
	  /* Add possible translation of cube */
	  xout[ic] += center[0];
	  yout[ic] += center[1];
	  zout[ic] += center[2];
	}
	/* Print line */
	fprintf(outFile,"Q");    /* Quad element */
	fprintf(outFile," %d",1);/* Boundary number 
				  * ("conducter number") */
	for (ic=0; ic<4; ic++){
	  iplot = ic;
	  if (swapNormal) iplot = 3-ic;
	  /* Print this corner to file */
	  fprintf(outFile," %16.8e %16.8e %16.8e",
		 xout[iplot],yout[iplot],zout[iplot]);
	}
	fprintf(outFile,"\n");
      }
    }
  }

  return 1;
}
