/*
Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA.
All rights reserved.

This Agreement gives you, the LICENSEE, certain rights and obligations.
By using the software, you indicate that you have read, understood, and
will comply with the terms.

Permission to use, copy and modify for internal, noncommercial purposes
is hereby granted.  Any distribution of this program or any part thereof
is strictly prohibited without prior written consent of M.I.T.

Title to copyright to this software and to any associated documentation
shall at all times remain with M.I.T. and LICENSEE agrees to preserve
same.  LICENSEE agrees not to make any copies except for LICENSEE'S
internal noncommercial use, or to use separately any portion of this
software without prior written consent of M.I.T.  LICENSEE agrees to
place the appropriate copyright notice on any such copies.

Nothing in this Agreement shall be construed as conferring rights to use
in advertising, publicity or otherwise any trademark or the name of
"Massachusetts Institute of Technology" or "M.I.T."

M.I.T. MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  By
way of example, but not limitation, M.I.T. MAKES NO REPRESENTATIONS OR
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR
THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS OR DOCUMENTATION WILL
NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
M.I.T. shall not be held liable for any liability nor for any direct,
indirect or consequential damages with respect to any claim by LICENSEE
or any third party on account of or arising from this Agreement or use
of this software.
*/
/* This program wil generate input files with constant-type 
 * elements (triangles) for a simple case.
 * At the moment most of the stuff is hardcoded. 
 * Eventually, most should be reached through command line arguments */

# include <stdlib.h>
# include <stdio.h>
# include <float.h>
# include <math.h>
# include <string.h>
# include <assert.h>

#define PI 3.141592653589793
#define ABS(A) ( ( (A) > 0 ) ? (A) : (-(A)) )

#define XPROD(V,V1,V2){ \
V[0] = V1[1]*V2[2]-V1[2]*V2[1]; \
V[1] = V1[2]*V2[0]-V1[0]*V2[2]; \
V[2] = V1[0]*V2[1]-V1[1]*V2[0];}

# define VNORM(V1) sqrt(V1[0]*V1[0]+V1[1]*V1[1]+V1[2]*V1[2])

/* Headers: */
void printTriangle(double *X0,double *X1,double *X2, 
		    int bcPart, FILE *outFile);
void refineTriangle(double *X0,double *X1,double *X2, 
		    double radius, 
		    double *center,
		    int bcPart, FILE *outFile, 
		    int level, int maxLevel);


int main(int argc, char *argv[])
{
  int nDense; /* Higher value = denser grid */
  int i, nn, dostats, swapNormal, bcPart;
  double radius, center[3], r;
  double area, minarea, maxarea, meanarea;
  double tetra0[3],tetra1[3],tetra2[3],tetra3[3];
  int neles;
  char inStr[]="inwards", outStr[]="outwards", *normalStr;
  FILE *outFile;

  outFile = fopen("sphere.qui","w");

  /* Generate a cube with equal size mesh and outward pointing normal */
  radius=1.0;
  center[0]=0.0;
  center[1]=0.0;
  center[2]=0.0;
  /* nDense by nDense elements on each side. */
  nDense = 6;
  /* Possibly swap normal to point inwards: */
  swapNormal=1;

  /* Set to 0 if no statistics are needed 
   * (written as comments at end of file) */
  dostats = 0;

  /* Expected number of elements: */
  nn = 4*nDense*nDense;

  normalStr = outStr;
  if (swapNormal) normalStr = inStr;
  /* Write header: */
  fprintf(outFile,
	  "0 Sphere w/ %s pointing normal. %d triangular elements.\n",
	  normalStr,nn);
  fprintf(outFile,
	  "* This file was automatically generated by \"spheregen2\"\n");

  /* Plot # elements to be generated: */
  fprintf(stdout,"* Generating %d elements\n",nn);

  /* Set up tetrahedron corners */
  tetra0[0] = sqrt(3.0)/3.0;
  tetra0[1] = 0.0;
  tetra0[2] = -sqrt(6.0)/12.0;

  tetra1[0] = -sqrt(3.0)/6.0;
  tetra1[1] = 0.5;
  tetra1[2] = -sqrt(6.0)/12.0;

  tetra2[0] = -sqrt(3.0)/6.0;
  tetra2[1] = -0.5;
  tetra2[2] = -sqrt(6.0)/12.0;

  tetra3[0] = 0.0;
  tetra3[1] = 0.0;
  tetra3[2] = sqrt(6.0)/4.0;

  /* Scale to radius of sphere */
  for (i=0; i<3; i++){
    tetra0[i] *= radius * 4.0 / sqrt(6.0);
    tetra1[i] *= radius * 4.0 / sqrt(6.0);
    tetra2[i] *= radius * 4.0 / sqrt(6.0);
    tetra3[i] *= radius * 4.0 / sqrt(6.0);
  }

  /* Test radius of tetrahedron */
  assert(ABS((r=VNORM(tetra0))-radius)<1.0e-8);
  assert(ABS((r=VNORM(tetra1))-radius)<1.0e-8);
  assert(ABS((r=VNORM(tetra2))-radius)<1.0e-8);
  assert(ABS((r=VNORM(tetra3))-radius)<1.0e-8);

  /* Start out with triangles: 
   * p0 p1 p2
   * p0 p3 p1
   * p0 p2 p3
   * p1 p3 p2  */

  /* Test output: 
  printTriangle(tetra0,tetra1,tetra2,1,outFile);
  printTriangle(tetra0,tetra3,tetra1,1,outFile);
  printTriangle(tetra0,tetra2,tetra3,1,outFile);
  printTriangle(tetra1,tetra3,tetra2,1,outFile);*/

  /* Refine each triangle in the tetrahedron */
  refineTriangle(tetra0,tetra1,tetra2,
		 radius, center,
		 bcPart=1, outFile, 1, nDense);
  refineTriangle(tetra0,tetra3,tetra1,
		 radius, center,
		 bcPart=1, outFile, 1, nDense);
  refineTriangle(tetra0,tetra2,tetra3, 
		 radius, center,
		 bcPart=1, outFile, 1, nDense);
  refineTriangle(tetra1,tetra3,tetra2, 
		 radius, center,
		 bcPart=1, outFile, 1, nDense);



  /* Initiate statistics settings? */
  minarea=DBL_MAX;
  maxarea=0.0;
  meanarea=0.0;
  neles=0;

  

  /* For statistics complete the calculations and write output */
  if (dostats){
    meanarea /= (double) neles;
    fprintf(stdout,"* Statistics:\n");
    fprintf(stdout,"*  %d elements\n",neles);
    fprintf(stdout,"*  Mean size %16.8e\n",meanarea);
    fprintf(stdout,"*  Minimum size %16.8e\n",minarea);
    fprintf(stdout,"*  Maximum size %16.8e\n",maxarea);
    fprintf(stdout,"*  Ratio between largest and smallest area %16.8e\n",
	    maxarea/minarea);
    area = 4.0*PI*radius*radius;
    fprintf(stdout,"*  Relative error on sphere area %16.8e\n",
	    (neles*meanarea-area)/area );
  }

  return 1;
}

void printTriangle(double *X0,double *X1,double *X2, 
		    int bcPart, FILE *outFile){
  int i,j;
  fprintf(outFile,"T");
  fprintf(outFile," %d",bcPart);
  fprintf(outFile," %16.8e %16.8e %16.8e",X0[0],X0[1],X0[2]);
  fprintf(outFile," %16.8e %16.8e %16.8e",X1[0],X1[1],X1[2]);
  fprintf(outFile," %16.8e %16.8e %16.8e",X2[0],X2[1],X2[2]);
  fprintf(outFile,"\n");
  /* Do statistics here */
  return;
}
void refineTriangle(double *X0,double *X1,double *X2, 
		    double radius, 
		    double *center,
		    int bcPart, FILE *outFile, 
		    int level, int maxLevel){
  int i,j;
  double S01[3],S12[3],S20[3], radS01,radS12,radS20;

  assert(level<=maxLevel);
  /* If we are at the finest level, print the triangle. */
  if (level==maxLevel){
    /* Add coordinates of center point */
    for (i=0;i<3;i++){
      X0[i] += center[i];
      X1[i] += center[i];
      X2[i] += center[i];
    }
    printTriangle(X0,X1,X2,bcPart,outFile);
  }
  /* Otherwise refine the triangle go to finer level */
  else {
    /* Find mid points of sides */
    for (i=0;i<3;i++){
      S01[i] = 0.5*(X0[i]+X1[i]);
      S12[i] = 0.5*(X1[i]+X2[i]);
      S20[i] = 0.5*(X2[i]+X0[i]);
    }
    /* Normalize sides with radius ("stretch" to be on sphere) */
    radS01 = VNORM(S01);
    radS12 = VNORM(S12);
    radS20 = VNORM(S20);
    for (i=0;i<3;i++){
      S01[i] /= radS01;
      S12[i] /= radS12;
      S20[i] /= radS20;
    }
    /* Refine using four new triangles (all ordered clockwise): 
     * S20-X0-S01, S01-X1-S12, S12-X2-S20, S02-S12-S20            */
    refineTriangle(S20,X0,S01,
		   radius,center,bcPart,outFile, 
		   level+1,maxLevel);
    refineTriangle(S01,X1,S12,
		   radius,center,bcPart,outFile, 
		   level+1,maxLevel);
    refineTriangle(S12,X2,S20,
		   radius,center,bcPart,outFile, 
		   level+1,maxLevel);
    refineTriangle(S01,S12,S20,
		   radius,center,bcPart,outFile, 
		   level+1,maxLevel);
  }
  return;
}

